import mulder
import numpy
from numpy.testing import assert_allclose


def test_frame():
    """Test local frames."""

    frame0 = mulder.LocalFrame()
    assert frame0.latitude == 0
    assert frame0.longitude == 0
    assert frame0.altitude == 0
    assert frame0.azimuth == 0
    assert frame0.elevation == 0

    frame1 = mulder.LocalFrame(altitude=1, azimuth=30)
    assert frame1.altitude == 1
    assert frame1.azimuth == 30

    ex = frame0.transform((1, 0, 0), destination=frame1, mode="vector")
    assert_allclose(ex, [numpy.sqrt(3) / 2, 0.5, 0.0], atol=1E-07)

    ex = frame0.transform((1, 0, 2), destination=frame1, mode="point")
    assert_allclose(ex, [numpy.sqrt(3) / 2, 0.5, 1.0], atol=1E-07)

    frame1 = mulder.LocalFrame(elevation=30)
    assert frame1.elevation == 30

    ex = frame0.transform((0, 1, 0), destination=frame1, mode="vector")
    assert_allclose(ex, [0.0, numpy.sqrt(3) / 2, -0.5], atol=1E-07)



def init():
    from .._core import picture
    import sys

    this = sys.modules[__name__]
    setattr(this, "__all__", picture.__all__)
    for name in picture.__all__:
        setattr(this, name, getattr(picture, name))

init()
del init



import mulder


def test_physics():
    """Test the physics interface."""

    physics = mulder.Physics(cutoff=1E-02)
    assert physics.bremsstrahlung == "SSR19"
    assert physics.pair_production == "SSR19"
    assert physics.photonuclear == "DRSS01"
    assert physics.cutoff == 1E-02
    assert physics.elastic_ratio == 5E-02

    physics.cutoff = 5E-2
    assert physics.cutoff == 5E-02

    rock = physics.compile("HumidRock")
    assert isinstance(rock.definition, mulder.materials.Composite)

    s = rock.stopping_power(1.0)
    assert s.size == 1
    assert s.shape == ()

    s = rock.stopping_power((1.0, 2.0))
    assert s.size == 2
    assert s.shape == (2,)



import mulder
from numpy.testing import assert_allclose
from pathlib import Path
import pytest


PREFIX = Path(__file__).parent


def test_earth():
    """Test Earth geometry."""

    # Test constructor & attributes.
    geometry = mulder.EarthGeometry(
        mulder.Layer(-1000, material="Rock"),
        mulder.Layer(0.0, material="Water"),
    )

    assert_allclose(geometry.zlim, [-11000, 0])

    assert len(geometry.layers) == 2
    assert geometry.layers[0].material == "Rock"
    assert geometry.layers[0].density == None
    assert geometry.layers[1].material == "Water"
    assert geometry.layers[1].density == None

    # Test the locate method.
    assert geometry.locate(
        position=[0.0, 0.0, -1001.0], frame=mulder.LocalFrame()
    ) == 0
    assert geometry.locate(altitude=-1001) == 0
    assert geometry.locate(
        position=[0.0, 0.0, -1.0], frame=mulder.LocalFrame()
    ) == 1
    assert geometry.locate(altitude=-1) == 1
    assert geometry.locate(
        position=[0.0, 0.0, -2.0], frame=mulder.LocalFrame(altitude=1)
    ) == 1
    assert geometry.locate(
        position=[0.0, 0.0, -0.5], frame=mulder.LocalFrame(altitude=1)
    ) == 2

    # Test the trace method.
    i = geometry.trace(altitude=-5, elevation=90)
    assert i["before"] == 1
    assert i["after"] == 2
    assert_allclose(i["distance"], 5)
    assert_allclose(i["altitude"], 0, atol=1E-06)

    i = geometry.trace(position=[0, 0, -5], direction=[0, 0, 1],
                       frame=mulder.LocalFrame())
    assert i["before"] == 1
    assert i["after"] == 2
    assert_allclose(i["distance"], 5)
    assert_allclose(i["position"], [0, 0, 0], atol=1E-06)

    i = geometry.trace(position=[0, 0, -6], direction=[0, 0, 1],
                       frame=mulder.LocalFrame(altitude=1))
    assert i["before"] == 1
    assert i["after"] == 2
    assert_allclose(i["distance"], 5)
    assert_allclose(i["position"], [0, 0, -1], atol=1E-06)

    i = geometry.trace(altitude=-1001, elevation=90)
    assert i["before"] == 0
    assert i["after"] == 1
    assert_allclose(i["distance"], 1, atol=1E-06)
    assert_allclose(i["altitude"], -1000, atol=1E-06)

    # Test the scan method.
    d = geometry.scan(altitude=-1001, elevation=90)
    assert_allclose(d, [1, 1000])

    d = geometry.scan(altitude=-1, elevation=90)
    assert_allclose(d, [0, 1])

    d = geometry.scan(altitude=1, elevation=-90)
    assert_allclose(d, [-geometry.zlim[0] - 1000, 1000])

    d = geometry.scan(position=[0, 0, -1001], direction=[0, 0, 1],
                      frame=mulder.LocalFrame())
    assert_allclose(d, [1, 1000], atol=1E-06)

    d = geometry.scan(position=[0, 0, -1002], direction=[0, 0, 1],
                      frame=mulder.LocalFrame(altitude=1))
    assert_allclose(d, [1, 1000], atol=1E-06)


@pytest.mark.requires_calzone
def test_local():
    """Test local geometry."""

    # Test constructor & attributes.
    geometry = mulder.LocalGeometry(PREFIX / "assets/geometry.toml")

    assert geometry.frame.latitude == 0
    assert geometry.frame.longitude == 0
    assert geometry.frame.altitude == 0
    assert geometry.frame.azimuth == 0
    assert geometry.frame.elevation == 0

    assert len(geometry.media) == 2
    assert geometry.media[0].material == "G4_AIR"
    assert geometry.media[0].description == "Environment"
    assert geometry.media[0].density == None
    assert geometry.media[1].material == "G4_CALCIUM_CARBONATE"
    assert geometry.media[1].description == "Environment.Ground"
    assert geometry.media[1].density == None

    frame = mulder.LocalFrame(latitude=37, longitude=3)
    geometry = mulder.LocalGeometry(
        PREFIX / "assets/geometry.toml", frame=frame
    )
    assert geometry.frame.latitude == 37
    assert geometry.frame.longitude == 3

    # Test the locate method.
    assert geometry.locate(position=[0.0, 0.0, 1.0]) == 0
    assert geometry.locate(position=[0.0, 0.0, -1.0]) == 1
    assert geometry.locate(latitude=37, longitude=3, altitude=-1.0) == 1
    media = geometry.locate(position=[
        [0.0, 0.0, -1001],
        [0.0, 0.0, -999],
        [0.0, 0.0, 999],
        [0.0, 0.0, 1001],
    ])
    assert_allclose(media, [2, 1, 0, 2])
    frame = mulder.LocalFrame(latitude=37, longitude=3, altitude=1)
    assert geometry.locate(position=[0.0, 0.0, -0.5], frame=frame) == 0
    assert geometry.locate(position=[0.0, 0.0, -1.5], frame=frame) == 1

    # Test the trace method.
    i = geometry.trace(
        position=[0, 0, -5],
        direction=[0, 0, 1],
    )
    assert i["before"] == 1
    assert i["after"] == 0
    assert i["distance"] == 5
    assert_allclose(i["position"], [0, 0, 0])

    i = geometry.trace(latitude=37, longitude=3, altitude=-5, elevation=90)
    assert i["before"] == 1
    assert i["after"] == 0
    assert_allclose(i["distance"], 5, atol=1E-07)
    assert_allclose(i["altitude"], 0, atol=1E-07)

    i = geometry.trace(position=[0, 0, -6], direction=[0, 0, 1], frame=frame)
    assert i["before"] == 1
    assert i["after"] == 0
    assert_allclose(i["distance"], 5, atol=1E-07)
    assert_allclose(i["position"], [0, 0, -1], atol=1E-07)

    i = geometry.trace(
        position=[0, 0, -1005],
        direction=[0, 0, 1],
    )
    assert i["before"] == 2
    assert i["after"] == 1
    assert i["distance"] == 5
    assert_allclose(i["position"], [0, 0, -1000])

    i = geometry.trace(
        position=[0, 0, 1005],
        direction=[0, 0, 1],
    )
    assert i["before"] == 2
    assert i["after"] == 2
    assert i["distance"] == 0
    assert_allclose(i["position"], [0, 0, 1005])

    # Test the scan method.
    d = geometry.scan(
        position=[0, 0, -995],
        direction=[0, 0, 1],
    )
    assert_allclose(d, [1000, 995])

    d = geometry.scan(
        position=[0, 0, -1005],
        direction=[0, 0, 1],
    )
    assert_allclose(d, [1000, 1000])

    d = geometry.scan(latitude=37, longitude=3, altitude=-995, elevation=90)
    assert_allclose(d, [1000, 995])

    d = geometry.scan(position=[0, 0, -996], direction=[0, 0, 1], frame=frame)
    assert_allclose(d, [1000, 995])

    n = geometry.media[0].normal(position=[0, 0, 1000])
    assert_allclose(n, [0, 0, 1])



import mulder
import os
from pathlib import Path


def test_config():
    """Test config data."""

    # Test the cache.
    HOME = os.getenv("HOME")
    default_cache = Path(HOME) / ".cache/mulder"
    assert isinstance(mulder.config.CACHE, Path)
    assert mulder.config.CACHE == default_cache

    cache = Path("/tmp/mulder")
    mulder.config.CACHE = cache
    assert mulder.config.CACHE == cache

    mulder.config.CACHE = default_cache

    # Test the notify flag.
    assert mulder.config.NOTIFY == True
    mulder.config.NOTIFY = False
    assert mulder.config.NOTIFY == False
    mulder.config.NOTIFY = True

    # Test static data.
    assert isinstance(mulder.config.PREFIX, Path)
    assert isinstance(mulder.config.VERSION, str)



def init():
    from .._core import materials
    import sys

    this = sys.modules[__name__]
    setattr(this, "__all__", materials.__all__)
    for name in materials.__all__:
        setattr(this, name, getattr(materials, name))

init()
del init



import mulder
import numpy
from numpy.testing import assert_allclose
from pathlib import Path


PREFIX = Path(__file__).parent


def test_constructor():
    """Test constructor function."""

    x = numpy.linspace(-1.5, 1.5, 4)
    y = numpy.linspace(-1.0, 1.0, 3)
    z = numpy.arange(12).reshape(3, 4)
    xlim = (x[0], x[-1])
    ylim = (y[0], y[-1])

    PARAMETERS = (
        (PREFIX / "assets/dem.asc", {}),
        (PREFIX / "assets/dem.grd", {}),
        (z, {"xlim": xlim, "ylim": ylim}),
    )

    CRSS = \
        list(range(27571, 27575)) + \
        [2154] + \
        [4326] + \
        list(range(32601, 32660)) + \
        list(range(32701, 32760))

    # Test loaders.
    for data, kwargs in PARAMETERS:
        grid = mulder.Grid(data, **kwargs)
        assert grid.xlim == (-1.5, 1.5)
        assert grid.ylim == (-1.0, 1.0)
        assert grid.zlim == (0.0, 11.0)
        assert grid.crs == 4326
        assert_allclose(grid.z(x, y), z, atol=1E-04)

        for crs in CRSS:
            grid = mulder.Grid(data, crs=crs, **kwargs)
            assert grid.crs == crs


def test_methods():
    """Test grid methods."""

    grid = mulder.Grid(PREFIX / "assets/dem.asc")
    assert grid.z(0.0, 0.0) == 5.5
    assert grid.z((0.0, 0.0)) == 5.5

    x = numpy.linspace(-1.5, 1.5, 4)
    y = numpy.linspace(-1.0, 1.0, 3)
    X, Y = numpy.meshgrid(x, y)
    xy = list(zip(X.flatten(), Y.flatten()))
    z0 = grid.z(xy)
    z1 = grid.z(x, y).flatten()
    assert_allclose(z0, z1)
    g0 = grid.gradient(xy)
    g1 = grid.gradient(x, y)
    assert g1.shape == (3, 4, 2)
    assert_allclose(g0, g1.reshape(g0.shape))

    x = numpy.linspace(-1.5, 1.5, 21)
    y = numpy.linspace(-1.0, 1.0, 21)
    xy = list(zip(x, y))
    z = grid.z(xy)
    assert z.shape == (21,)
    for i, (x, y) in enumerate(xy):
        zi = grid.z(x, y)
        assert zi == grid.z((x, y))
        assert zi == z[i]
        assert_allclose(grid.gradient(x, y), grid.gradient((x, y)))

    dz = grid.gradient(0.0, 0.0)
    assert_allclose(dz, (1.0, 4.0), atol=1E-03)



from datetime import date
import mulder
import numpy
from numpy.testing import assert_allclose
from pathlib import Path


def test_constructor():
    """Test constructor function."""

    # Test default constructor.
    geomagnet = mulder.EarthMagnet()
    assert isinstance(geomagnet.date, date)
    assert str(geomagnet.date) == "2025-06-21"
    assert isinstance(geomagnet.zlim, tuple)
    assert_allclose(geomagnet.zlim, (-1E+03, 6E+05))
    assert geomagnet.model == "IGRF14"

    # Test arguments.
    geomagnet = mulder.EarthMagnet(
        Path(mulder.config.PREFIX) / "data/magnet/IGRF14.COF",
        date = "1978-08-16"
    )
    assert isinstance(geomagnet.date, date)
    assert str(geomagnet.date) == "1978-08-16"
    assert isinstance(geomagnet.zlim, tuple)
    assert_allclose(geomagnet.zlim, (-1E+03, 6E+05))
    assert geomagnet.model == "IGRF14"


def test_field():
    """Test field method."""

    latitude, longitude = 45.8, 3.1
    position = [1E+04, -1E+04, 1E+03]
    geomagnet = mulder.EarthMagnet()

    # Compute field in local frame.
    frame0 = mulder.LocalFrame(latitude=latitude, longitude=longitude)
    field0 = geomagnet.field(frame=frame0, position=position)

    # Compute field at the same point but using geographic coordinates.
    state1 = mulder.LocalStates(frame=frame0, position=position) \
        .to_geographic()
    field1 = geomagnet.field(state1)

    # Transform the latter field to the former local frame.
    state1.azimuth = 0.0
    state1.elevation = 0.0
    frame1 = mulder.LocalFrame(state1)
    field1 = frame1.transform(field1, destination=frame0, mode="vector")

    # Compare results.
    assert_allclose(field0, field1)



import mulder


def test_Random():
    """Test the Random interface."""

    rng = mulder.Random()
    assert rng.index == 0

    rng = mulder.Random(1)
    assert rng.seed == 1
    assert rng.index == 0
    v = rng.uniform01(10)

    rng = mulder.Random(1, index=9)
    assert rng.seed == 1
    assert rng.index == 9
    assert rng.uniform01() == v[-1]

    rng.seed = 0
    assert (rng.uniform01(10) == v).all()

    rng.index = 5
    assert (rng.uniform01(5) == v[5:]).all()



import mulder
import numpy
from numpy.testing import assert_allclose
from pathlib import Path


PREFIX = Path(__file__).parent


def test_constructor():
    """Test constructor function."""

    layer = mulder.Layer(PREFIX / "assets/dem.asc")
    assert len(layer.data) == 1
    assert isinstance(layer.data[0], mulder.Grid)
    assert layer.density == None
    assert layer.description == None
    assert layer.material == "Rock"

    layer = mulder.Layer(1.0, description="toto")
    assert len(layer.data) == 1
    assert layer.data[0] == 1.0
    assert layer.description == "toto"

    layer = mulder.Layer(PREFIX / "assets/dem.asc", 1.0, material="Water",
                         density=1.0)
    assert len(layer.data) == 2
    assert isinstance(layer.data[0], mulder.Grid)
    assert layer.data[1] == 1.0
    assert layer.material == "Water"
    assert layer.density == 1.0


def test_methods():
    """Test grid methods."""

    layer = mulder.Layer(PREFIX / "assets/dem.asc")
    assert layer.altitude(0.0, 0.0) == 5.5
    assert layer.altitude((0.0, 0.0)) == 5.5
    assert_allclose(layer.altitude(1.0, 0.0), 9.5, atol=1E-03)
    assert_allclose(layer.altitude(0.0, 1.0), 6.5, atol=1E-03)
    assert numpy.isnan(layer.altitude((2.0, 0.0)))
    assert (layer.normal((2.0, 0.0)) == 0.0).all()

    layer = mulder.Layer(PREFIX / "assets/dem.asc", 0.0)
    assert layer.altitude(0.0, 0.0) == 5.5
    assert layer.altitude((2.0, 0.0)) == 0.0
    assert (layer.normal((2.0, 0.0)) != 0.0).any()

    layer = mulder.Layer(1.0)
    assert layer.altitude(0.0, 0.0) == 1.0
    assert_allclose(layer.normal((0.0, 0.0)), (1.0, 0.0, 0.0), atol=1E-06)



import argparse
import os

import mulder


def main():
    """Entry point for the CLI."""

    parser = argparse.ArgumentParser(
        prog = "python3 -m mulder",
        description = "Command-line utility for Mulder.",
        epilog = "Copyright (C) Université Clermont Auvergne, CNRS/IN2P3, LPCA"
    )
    parser.add_argument("-c", "--cache",
        help = "Mulder default cache location.",
        action = "store_true"
    )
    parser.add_argument("-p", "--prefix",
        help = "Mulder installation prefix.",
        action = "store_true"
    )
    parser.add_argument("-v", "--version",
        help = "Mulder version.",
        action = "store_true"
    )

    args = parser.parse_args()

    result = []
    if args.cache:
        result.append(str(mulder.config.DEFAULT_CACHE))
    if args.prefix:
        result.append(str(mulder.config.PREFIX))
    if args.version:
        result.append(mulder.config.VERSION)

    if result:
        print(" ".join(result))
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



import mulder
import numpy
from numpy.testing import assert_allclose


MODELS = (
    "midlatitude-summer",
    "midlatitude-winter",
    "subartic-summer",
    "subartic-winter",
    "tropical",
    "us-standard"
)

TABLE = numpy.array((
    (      0, 1E+00),
    ( 10_000, 1E-01),
    (100_000, 1E-02),
))

def test_constructor():
    """Test constructor function."""

    # Test default constructor.
    atmosphere = mulder.Atmosphere()
    assert atmosphere.material == "Air"
    assert atmosphere.model == "us-standard"

    # Test predefined models.
    for model in MODELS:
        atmosphere = mulder.Atmosphere(model)
        assert atmosphere.material == "Air"
        assert atmosphere.model == model

    # Test table case.
    atmosphere = mulder.Atmosphere(TABLE)
    assert atmosphere.material == "Air"
    assert_allclose(atmosphere.model, TABLE)
    assert atmosphere.model.flags.writeable == False

    # Test material case.
    atmosphere = mulder.Atmosphere(MODELS[0], material="N2")
    assert atmosphere.material == "N2"
    assert atmosphere.model == MODELS[0]

    atmosphere = mulder.Atmosphere(TABLE, material="N2")
    assert atmosphere.material == "N2"
    assert_allclose(atmosphere.model, TABLE)
    assert atmosphere.model.flags.writeable == False


def test_density():
    """Test density method."""

    atmosphere = mulder.Atmosphere(TABLE)
    assert_allclose(atmosphere.density(TABLE[:,0]), TABLE[:,1])
    for z, rho in TABLE:
        assert atmosphere.density(z) == rho

    for i in range(len(TABLE) - 1):
        z = 0.5 * (TABLE[i + 1, 0] + TABLE[i, 0])
        assert atmosphere.density(z) < TABLE[i, 1]
        assert atmosphere.density(z) > TABLE[i + 1, 1]



import datetime
import mulder
import numpy
from numpy.testing import assert_allclose
from pathlib import Path
import pytest


PREFIX = Path(__file__).parent


def test_constructor():
    """Test the constructor."""

    meter = mulder.Fluxmeter()
    assert isinstance(meter.geometry, mulder.EarthGeometry)
    assert len(meter.geometry.layers) == 0
    assert meter.atmosphere.material == "Air"

    meter = mulder.Fluxmeter(1.0)
    assert isinstance(meter.geometry, mulder.EarthGeometry)
    assert len(meter.geometry.layers) == 1
    assert meter.geometry.layers[0].material == "Rock"
    assert meter.geometry.layers[0].data == (1.0,)
    assert meter.atmosphere.material == "Air"

    meter = mulder.Fluxmeter(geometry=PREFIX / "assets/geometry.toml")
    assert isinstance(meter.geometry, mulder.LocalGeometry)

    meter = mulder.Fluxmeter(
        atmosphere = "midlatitude-winter",
        date = "2025-12-25",
        mode = "mixed",
        bremsstrahlung = "KKP95",
        seed = 123456,
        reference = "Gaisser90",
    )

    assert meter.atmosphere.model == "midlatitude-winter"
    assert meter.geomagnet.date == datetime.date(2025, 12, 25)
    assert meter.mode == "mixed"
    assert meter.physics.bremsstrahlung == "KKP95"
    assert meter.random.seed == 123456
    assert meter.reference.model == "Gaisser90"

    meter = mulder.Fluxmeter(reference = 1)
    assert meter.reference.model == 1

    with pytest.raises(TypeError) as e:
        meter = mulder.Fluxmeter(
            geomagnet=mulder.EarthMagnet(),
            date="2025-12-25"
        )
    meter = mulder.Fluxmeter(geomagnet=mulder.EarthMagnet(date="2025-12-25"))
    assert meter.geomagnet.date == datetime.date(2025, 12, 25)

    with pytest.raises(TypeError) as e:
        meter = mulder.Fluxmeter(
            physics=mulder.Physics(),
            bremsstrahlung="KKP95"
        )
    meter = mulder.Fluxmeter(physics=mulder.Physics(bremsstrahlung="KKP95"))
    assert meter.physics.bremsstrahlung == "KKP95"

    with pytest.raises(TypeError) as e:
        meter = mulder.Fluxmeter(
            random=mulder.Random(),
            seed=123456
        )
    meter = mulder.Fluxmeter(random=mulder.Random(seed=123456))
    assert meter.random.seed == 123456


def test_attributes():
    """Test the attributes."""

    meter = mulder.Fluxmeter()

    meter.atmosphere = "midlatitude-winter"
    assert meter.atmosphere.model == "midlatitude-winter"

    meter.atmosphere = mulder.Atmosphere("midlatitude-winter")
    assert meter.atmosphere.model == "midlatitude-winter"


def test_continuous():
    """Test continuous mode."""

    meter = mulder.Fluxmeter(0.0)
    assert meter.mode == "continuous"

    s1 = dict(altitude=-30.0, elevation=70.0)
    s0 = meter.transport(**s1)
    assert_allclose(s0.altitude, meter.reference.altitude)

    f1a = meter.reference.flux(s0) * s0.weight
    f1b = meter.flux(**s1)
    assert_allclose(f1a, f1b)

    meter = mulder.Fluxmeter(0.0, geomagnet=True)

    s1a = dict(altitude=-30.0, elevation=70.0, pid=13)
    s0a = meter.transport(**s1a)
    s1b = dict(altitude=-30.0, elevation=70.0, pid=-13)
    s0b = meter.transport(**s1b)
    f1a = meter.reference.flux(s0a) * s0a.weight + \
          meter.reference.flux(s0b) * s0b.weight
    f1b = meter.flux(**s1)
    assert_allclose(f1a, f1b)


def test_stochastic():
    """Test stochastic modes."""

    for mode in ("discrete", "mixed"):
        meter = mulder.Fluxmeter(0.0, mode=mode, seed=123456)
        assert meter.mode == mode

        s1 = dict(altitude=-30.0, elevation=70.0)
        N = 1000
        s0 = meter.transport(**s1, events=N)
        assert_allclose(s0.altitude, meter.reference.altitude)

        f1a = sum(meter.reference.flux(s0) * s0.weight) / N
        f1b, s1b = meter.flux(**s1, events=N)
        assert_allclose(f1a, f1b, rtol=5E-02)



def init():
    from . import _core
    import sys

    for submodule in ("materials", "picture"):
        _core.__all__.remove(submodule)

    this = sys.modules[__name__]
    setattr(this, "__all__", _core.__all__)
    for name in _core.__all__:
        setattr(this, name, getattr(_core, name))

init()
del init



import mulder
import numpy
from numpy.testing import assert_allclose
import pytest


STATES = (
    mulder.GeographicStates,
    mulder.LocalStates,
)


def test_constructors():
    """Test constructor functions."""

    for States in STATES:
        for meth in ("empty", "full", "zeros"):
            shape = []
            a = getattr(States, meth)(shape)
            assert a.shape == tuple(shape)
            assert a.size == 1
            assert a.ndim == 0
            assert a.pid == None

            shape = 3
            if meth == "full":
                fill_value = { "energy": 1.0, "pid": 13 }
                a = getattr(States, "full")(shape, **fill_value)
            else:
                a = getattr(States, meth)(shape)
            assert a.shape == (shape,)
            assert a.size == 3
            assert a.ndim == 1

            if meth == "full":
                for k, v in fill_value.items():
                    assert numpy.all(a[k] == v)

                b = getattr(States, "from_array")(a.array)
                assert numpy.all(a == b)
                a["energy"] = 0
                assert numpy.all(a != b)

                b = getattr(States, "from_array")(a.array, copy=False)
                assert numpy.all(a == b)
                a["energy"] = 0
                assert numpy.all(a == b)

            elif meth == "zeros":
                for ai in a:
                    assert isinstance(ai, States)
                    for aij in ai:
                        assert aij == 0

            shape = (1, 2, 3)
            a = getattr(States, meth)(shape)
            assert a.shape == shape
            assert a.size == 6
            assert a.ndim == 3

        energy = (1, 2, 3)
        a = States(energy=energy)
        assert a.size == 3
        assert a.shape == (3,)
        assert a.ndim == 1
        assert numpy.all(a["energy"] == energy)

        a = States.zeros()
        assert a.dtype() == a.array.dtype

        a = States.zeros(tagged=True)
        assert numpy.all(a["pid"] == 0)
        assert a.dtype(tagged=True) == a.array.dtype


def test_attributes():
    """Test attributes interface."""

    for States in STATES:
        if States == mulder.GeographicStates:
            attr = {
                "energy": (1, 2, 3),
                "latitude": (45, 48, 51),
                "longitude": (3, 5, 7),
                "altitude": (4, 5, 6),
                "azimuth": (-2, -3, -4),
                "elevation": (8, 9, 10),
                "weight": (3.6, 7.2, 5.9),
            }
        else:
            attr = {
                "energy": (1, 2, 3),
                "position": ((1, 2, 3), (4, 5, 6), (7, 8, 9)),
                "direction": ((-1, -2, -3), (-4, -5, -6), (-7, -8, -9)),
                "weight": (3.6, 7.2, 5.9),
            }

        states = States(**attr)
        for k, v in attr.items():
            assert numpy.all(states[k] == v)
            assert numpy.all(getattr(states,k) == v)
            states[k] = 0
            assert numpy.all(getattr(states,k) == 0)
            setattr(states, k, 1)
            assert numpy.all(states[k] == 1)

        with pytest.raises(ValueError):
            assert states["pid"] == None
        assert states.pid == None

        pid = (13, -13, 13)
        states = States(pid=pid)
        assert numpy.all(states["pid"] == pid)
        assert numpy.all(states.pid == pid)
        states["pid"] = 0
        assert numpy.all(states.pid == 0)
        states.pid = 1
        assert numpy.all(states["pid"] == 1)

    frame = mulder.LocalFrame(latitude=-33, longitude=65)
    states = mulder.LocalStates(frame=frame, **attr)
    assert states.frame == frame
    with pytest.raises(ValueError):
        states["frame"]


def test_conversions():
    """Test conversions between states."""

    def normed(v):
        return numpy.array(v) / numpy.linalg.norm(v)

    a = mulder.LocalStates(energy=(1, 2, 3), direction=normed((1, 0, 1)))
    b = a.to_geographic()
    assert_allclose(b.latitude, 0.0)
    assert_allclose(b.longitude, 0.0)
    assert_allclose(b.altitude, 0.0, atol=1E-07)
    assert_allclose(b.azimuth, 90.0)
    assert_allclose(b.elevation, 45.0)
    b = b.to_local(a.frame)
    assert_allclose(a.direction, b.direction, atol=1E-07)

    a = mulder.LocalStates(direction=normed((0, 1, 1)))
    b = a.to_geographic()
    assert_allclose(b.azimuth, 0.0, atol=1E-07)
    assert_allclose(b.elevation, 45.0)
    b = b.to_local(a.frame)
    assert_allclose(a.direction, b.direction, atol=1E-07)

    a = mulder.LocalStates(direction=normed((1, 1, 0)))
    b = a.to_geographic()
    assert_allclose(b.azimuth, 45.0)
    assert_allclose(b.elevation, 0.0, atol=1E-07)
    b = b.to_local(a.frame)
    assert_allclose(a.direction, b.direction, atol=1E-07)

    frame = mulder.LocalFrame(
        latitude=-33,
        longitude=68,
        altitude=102,
        azimuth=28,
        elevation=-15,
    )
    a = mulder.LocalStates(
        frame=frame,
        position=(1, 2, 3),
        direction=normed((3, 2, 1))
    )
    b = a.to_geographic().to_local(a.frame)
    assert_allclose(a.position, b.position, atol=1E-07)
    assert_allclose(a.direction, b.direction, atol=1E-07)

    a = mulder.LocalStates(position=(0, 0, 100))
    b = mulder.GeographicStates.from_local(a)
    assert_allclose(a.position[2], b.altitude)

    frame1 = mulder.LocalFrame()
    a0 = mulder.LocalStates(
        frame=frame,
        position=(1, 2, 3),
        direction=normed((3, 2, 1))
    )
    a1 = a0.transform(destination=frame1)
    b1 = a0.to_geographic().to_local(frame1)
    assert_allclose(a1.position, b1.position)
    assert_allclose(a1.direction, b1.direction)



import mulder
import numpy
from numpy.testing import assert_allclose
import pytest


DEFAULT_RATIO = 1.2766

TABLE = numpy.zeros((2, 2, 2))
TABLE[0,0,:] = (0.5, 1.0)
TABLE[0,1,:] = (1.0, 2.0)
TABLE[1,0,:] = (2.0, 4.0)
TABLE[1,1,:] = (4.0, 8.0)


def test_parametric():
    """Test parametric models."""

    models = ("GCCLY15", "Gaisser90")
    for model in models:
        reference = mulder.Reference(model)
        assert_allclose(reference.energy, (1E-03, 1E+12))
        assert_allclose(reference.altitude, 0)
        assert_allclose(reference.elevation, (0, 90))

        f0 = reference.flux()
        f1 = reference.flux(pid=13)
        f2 = reference.flux(pid=-13)
        assert_allclose(f0, f1 + f2)
        assert_allclose(f2 / f1, DEFAULT_RATIO)

        with pytest.raises(TypeError) as e:
            reference = mulder.Reference(model, altitude=5)
        assert 'altitude' in str(e.value)


def test_flat():
    """Test flat model."""

    reference = mulder.Reference(2)
    assert_allclose(reference.energy, (1E-03, 1E+12))
    assert_allclose(reference.altitude, (-11E+03, 120E+03))
    assert_allclose(reference.elevation, (-90, 90))

    f0 = reference.flux()
    f1 = reference.flux(pid=13)
    f2 = reference.flux(pid=-13)
    assert_allclose(f0, 2.0)
    assert_allclose(f0, f1 + f2)
    assert_allclose(f2 / f1, DEFAULT_RATIO)


def test_tabulated():
    """Test tabulated model."""

    ALTITUDE = (-1, 1)
    TEST_CASES = (
        ((2, 2), None, 1.0, DEFAULT_RATIO),
        ((2, 2, 2), None, 2.0, 1.0),
        ((2, 2, 2), ALTITUDE, 1.0, DEFAULT_RATIO),
        ((2, 2, 2, 2), ALTITUDE, 2.0, 1.0),
    )

    for (shape, altitude, expected, ratio) in TEST_CASES:
        reference = mulder.Reference(
            numpy.ones(shape),
            energy=(0.1, 10),
            altitude=altitude,
        )
        f0 = reference.flux()
        f1 = reference.flux(pid=13)
        f2 = reference.flux(pid=-13)
        assert_allclose(f0, expected, atol=1E-07)
        assert_allclose(f0, f1 + f2, atol=1E-07)
        assert_allclose(f2 / f1, ratio, atol=1E-04)

    energy = numpy.geomspace(1E-03, 1E+12, 151)
    cos_theta = numpy.linspace(0, 1.0, 51)
    X, Y = numpy.meshgrid(energy, cos_theta)
    reference0 = mulder.Reference("GCCLY15")
    elevation = 90.0 - numpy.degrees(numpy.arccos(Y[:]))
    flux = reference0.flux(energy=X[:], elevation=elevation)
    table = numpy.empty((51, 151, 2))
    table[:,:,0] = flux / (1 + DEFAULT_RATIO)
    table[:,:,1] = flux * DEFAULT_RATIO / (1 + DEFAULT_RATIO)

    reference1 = mulder.Reference(
        table,
        energy=(energy[0], energy[-1]),
        cos_theta=(cos_theta[0], cos_theta[-1]),
    )

    assert_allclose(reference1.energy, (energy[0], energy[-1]))
    assert_allclose(reference1.elevation, (0.0, 90.0))
    assert reference1.altitude == 0.0
    assert_allclose(reference1.model, table, atol=1E-07)

    TEST_VALUES = [
        (1.5E+00, 47),
        (3.0E+02, 15),
        (5.0E+06, 87),
        (2.0E+09, 55),
    ]

    for (energy, elevation) in TEST_VALUES:
        kwargs = dict(energy=energy, elevation=elevation)
        f0 = reference0.flux(**kwargs)
        f1 = reference1.flux(**kwargs)
        f2 = reference1.flux(**kwargs, pid=13)
        f3 = reference1.flux(**kwargs, pid=-13)
        assert_allclose(f0, f1, rtol=1E-02)
        assert_allclose(f1, f2 + f3)
        assert_allclose(f3 / f2, DEFAULT_RATIO, atol=1E-04)



import mulder
import numpy
from numpy.testing import assert_allclose
from pathlib import Path


PREFIX = Path(__file__).parent


def test_camera():
    """Test camera object."""

    frame = mulder.LocalFrame(azimuth=45, elevation=30)
    camera = frame.camera()
    assert camera.frame == frame
    assert camera.fov == 60
    assert camera.ratio == 4 / 3
    assert camera.resolution == (90, 120)
    assert_allclose(camera.focal, 1.732051, atol=1E-05)

    camera = frame.camera(focal=1)
    assert_allclose(camera.focal, 1)
    assert_allclose(camera.fov, 90.0, atol=1E-05)

    camera = frame.camera((18, 32))
    assert camera.ratio == 16 / 9
    assert camera.resolution == (18, 32)
    assert_allclose(camera.focal, 1.732051, atol=1E-05)


def test_projection():
    """Test projection object."""

    geometry = mulder.EarthGeometry(0)
    frame = mulder.LocalFrame(altitude=1)
    camera = frame.camera((9, 13))
    projection = camera.project(geometry)
    assert projection.frame == frame
    assert projection.medium.shape == (9, 13)
    expected = numpy.full((9, 13), 0, dtype="i4")
    expected[4:,:] = 1
    assert_allclose(projection.medium, expected)
    expected = numpy.full((9, 13), 0, dtype="i4")
    expected[4:,:] = 120E+03
    assert_allclose(projection.altitude, expected, atol=1E-04)

    geometry = mulder.LocalGeometry(PREFIX / "assets/geometry.toml")
    frame = mulder.LocalFrame(altitude=1)
    camera = frame.camera((9, 13))
    projection = camera.project(geometry)
    assert projection.frame == frame
    assert projection.medium.shape == (9, 13)
    expected = numpy.full((9, 13), 1, dtype="i4")
    expected[4:,:] = 0
    assert_allclose(projection.medium, expected)


def test_pixels():
    """Test pixels object."""

    frame = mulder.LocalFrame(azimuth=45, elevation=30)
    camera = frame.camera((9, 13))
    pixels = camera.pixels
    assert pixels.u.shape == (camera.resolution[1],)
    assert pixels.v.shape == (camera.resolution[0],)
    assert pixels.u[6] == 0.5
    assert pixels.v[4] == 0.5
    assert pixels.azimuth.shape == camera.resolution
    assert_allclose(pixels.azimuth[4, 6], frame.azimuth, atol=1E-07)
    assert pixels.elevation.shape == camera.resolution
    assert_allclose(pixels.elevation[4, 6], frame.elevation, atol=1E-07)



import mulder.materials as materials
from numpy.testing import assert_allclose
import os
from pathlib import Path
from tempfile import TemporaryDirectory


PREFIX = Path(__file__).parent


def test_composite():
    """Test the composite interface."""

    composite = materials.Composite.define(
        "Composite", composition=("Rock", "Water")
    )
    assert composite.composition == ("Rock", "Water")
    assert_allclose(composite.density, 1473.02, atol=0.01)

    composite = materials.Composite.define("Composite",
        composition=(("Rock", 1.0), ("Water", 0.0))
    )
    assert composite.density == 2.65E+03

    composite["Rock"] = 0.0
    composite["Water"] = 1.0
    assert composite.density == 1.02E+03

    composites = materials.Composite.all()
    assert "Composite" in composites


def test_element():
    """Test the element interface."""

    H = materials.Element("H")
    assert H.Z == 1
    assert H.A == 1.008
    assert H.I == 19.2E-09

    U_238 = materials.Element.define("U-238", Z=92, A=238.0508, I=890E-09)
    assert U_238.Z == 92
    assert U_238.A == 238.0508
    assert_allclose(U_238.I, 890E-09)

    elements = materials.Element.all()
    assert "U-238" in elements


def test_base():
    """Test the base interface."""

    rock = materials.Material("Rock")
    assert rock.composition == (("Rk", 1.0),)
    assert rock.density == 2.65E+03
    assert rock.I == None

    water = materials.Material("Water")
    assert water.composition[0][0] == "H"
    assert water.composition[1][0] == "O"
    assert_allclose(
        [c[1] for c in water.composition],
        [0.111894, 0.888106],
        atol = 1E-04
    )
    assert water.density == 1.02E+03
    assert water.I == None

    air = materials.Material("Air")
    assert [c[0] for c in air.composition] == ["Ar", "C", "N", "O"]
    assert air.density == 1.205
    assert air.I == None

    base = materials.Material.define(
        "TestRock",
        composition = {
            "Rock": 0.95,
            "Ar": 0.05,
        },
        density = 2.0E+03,
        I = 120E-09
    )
    assert base.composition[0][0] == "Ar"
    assert base.composition[1][0] == "Rk"
    assert_allclose(
        [c[1] for c in base.composition],
        [0.05, 0.95]
    )
    assert base.density == 2E+03
    assert base.I == 120E-09

    base2 = materials.Material.define(
        "TestRock2",
        composition = (
            ("Rock", 0.95),
            ("Ar", 0.05),
        ),
        density = 2.0E+03,
        I = 120E-09
    )
    assert base.composition == base2.composition

    H2O = materials.Material.define(
        "PureWater", composition="H2O", density=1E+03
    )
    assert H2O.composition == water.composition

    bases = materials.Material.all()
    assert "PureWater" in bases


def test_dump():
    """Test the dump function."""

    cwd = os.getcwd()
    with TemporaryDirectory() as d:
        os.chdir(d)
        materials.dump("materials.toml")
        materials.load("materials.toml")
        os.chdir(cwd)  # Must be before tmpdir removal, on Windows.


def test_load():
    """Test the load function."""

    materials.load(PREFIX / "assets/materials.toml")

    base = materials.Material("MoistAir")
    assert [c[0] for c in base.composition] == ["Ar", "C", "H", "N", "O"]
    assert base.density == 1.2

    composite = materials.Composite("HumidRock")
    assert composite.composition == ("Rock", "Water")
